from django.urls import reverse
from rest_framework import status

from arkindex.images.models import ImageServer
from arkindex.project.tests import FixtureTestCase


class TestCustomAdminViews(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.root_srv = ImageServer.objects.create(display_name='Root', url='http://server.dev.teklia.com/iiif')
        cls.sub_srv = ImageServer.objects.create(display_name='Subfolder', url='http://server.dev.teklia.com/iiif/sub')

        cls.root_srv.images.create(path='some/thing', width=100, height=100)
        cls.sub_srv.images.create(path='other/thing', width=100, height=100)

    def test_merge_form_requires_login(self):
        url = reverse('admin:images_imageserver_merge', kwargs={'id': str(self.sub_srv.id)})
        response = self.client.get(url)
        self.assertRedirects(response, '{}?next={}'.format(reverse('admin:login'), url))

    def test_merge_form_requires_admin(self):
        self.client.force_login(self.user)
        url = reverse('admin:images_imageserver_merge', kwargs={'id': str(self.sub_srv.id)})
        response = self.client.get(url)
        self.assertRedirects(response, '{}?next={}'.format(reverse('admin:login'), url))

    def test_merge_form(self):
        """
        Test the merge form holds the proper context objects for the admin templates
        """
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('admin:images_imageserver_merge', kwargs={'id': str(self.sub_srv.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.context.get('object'), self.sub_srv)
        self.assertEqual(response.context.get('opts'), ImageServer._meta)

    def test_perform_merge(self):
        self.assertTrue(ImageServer.objects.filter(id=self.sub_srv.id).exists())
        self.assertEqual(self.root_srv.images.count(), 1)

        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('admin:images_imageserver_merge', kwargs={'id': str(self.sub_srv.id)}),
            {'destination': str(self.root_srv.id)},
            follow=True,
        )
        self.assertRedirects(response, reverse('admin:images_imageserver_changelist'))

        # Success notification
        message = next(iter(response.context['messages']))
        self.assertEqual(message.tags, "success")
        self.assertEqual(message.message, 'Successfully merged Subfolder into Root')

        self.assertFalse(ImageServer.objects.filter(id=self.sub_srv.id).exists())
        self.assertEqual(self.root_srv.images.count(), 2)

    def test_merge_error(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('admin:images_imageserver_merge', kwargs={'id': str(self.root_srv.id)}),
            {'destination': str(self.sub_srv.id)},
            follow=True,
        )
        self.assertRedirects(response, reverse('admin:images_imageserver_changelist'))

        # No change
        self.assertTrue(ImageServer.objects.filter(id=self.sub_srv.id).exists())
        self.assertEqual(self.root_srv.images.count(), 1)
        self.assertEqual(self.sub_srv.images.count(), 1)

        # Error message
        message = next(iter(response.context['messages']))
        self.assertEqual(message.tags, "error")
        self.assertEqual(
            message.message,
            'Merging failed: Cannot merge into an image server on a child folder: '
            'Some images are not in the child folder',
        )

    def test_split_form_requires_login(self):
        url = reverse('admin:images_imageserver_split', kwargs={'id': str(self.root_srv.id)})
        response = self.client.get(url)
        self.assertRedirects(response, '{}?next={}'.format(reverse('admin:login'), url))

    def test_split_form_requires_admin(self):
        self.client.force_login(self.user)
        url = reverse('admin:images_imageserver_split', kwargs={'id': str(self.root_srv.id)})
        response = self.client.get(url)
        self.assertRedirects(response, '{}?next={}'.format(reverse('admin:login'), url))

    def test_perform_split(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('admin:images_imageserver_split', kwargs={'id': str(self.root_srv.id)}),
            {'folder': 'some'},
            follow=True,
        )
        self.assertRedirects(response, reverse('admin:images_imageserver_changelist'))

        self.root_srv.refresh_from_db()
        self.assertEqual(self.root_srv.images.count(), 0)
        new_srv = ImageServer.objects.get(url='http://server.dev.teklia.com/iiif/some')
        self.assertEqual(new_srv.display_name, 'Root_some')
        self.assertEqual(new_srv.images.count(), 1)

        # Success notification
        message = next(iter(response.context['messages']))
        self.assertEqual(message.tags, "success")
        self.assertEqual(message.message, 'Successfully split "some" from Root into Root_some')

    def test_split_error(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('admin:images_imageserver_split', kwargs={'id': str(self.root_srv.id)}),
            {'folder': 'dev/null'},
            follow=True,
        )
        self.assertRedirects(response, reverse('admin:images_imageserver_changelist'))

        # No change
        self.root_srv.refresh_from_db()
        self.sub_srv.refresh_from_db()
        self.assertEqual(self.root_srv.images.count(), 1)
        self.assertEqual(self.sub_srv.images.count(), 1)

        # Error message
        message = next(iter(response.context['messages']))
        self.assertEqual(message.tags, "error")
        self.assertEqual(
            message.message,
            'Splitting failed: No images are in the "dev/null" folder',
        )
