import hashlib
import uuid
from collections import defaultdict
from textwrap import dedent

from django.db.models import Count, Q
from django.shortcuts import get_object_or_404, redirect
from django.utils import timezone
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, extend_schema, extend_schema_view
from rest_framework.authentication import SessionAuthentication, TokenAuthentication
from rest_framework.exceptions import ValidationError
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveUpdateAPIView,
    UpdateAPIView,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from arkindex.ponos.authentication import AgentAuthentication, TaskAuthentication
from arkindex.ponos.keys import load_private_key
from arkindex.ponos.models import Agent, Artifact, Farm, Secret, State, Task
from arkindex.ponos.permissions import (
    IsAgent,
    IsAgentOrArtifactGuest,
    IsAgentOrTaskGuest,
    IsAssignedAgentOrReadOnly,
    IsAssignedAgentOrTaskOrReadOnly,
    IsTask,
    IsTaskAdmin,
)
from arkindex.ponos.renderers import PublicKeyPEMRenderer
from arkindex.ponos.serializers import (
    AgentActionsSerializer,
    AgentCreateSerializer,
    AgentDetailsSerializer,
    AgentLightSerializer,
    AgentStateSerializer,
    ArtifactSerializer,
    ClearTextSecretSerializer,
    FarmSerializer,
    NewTaskSerializer,
    TaskDefinitionSerializer,
    TaskSerializer,
    TaskTinySerializer,
)
from arkindex.project.permissions import IsVerified
from rest_framework_simplejwt.views import TokenRefreshView


class PublicKeyEndpoint(APIView):
    """
    Fetch the server's public key in PEM format to perform agent registration.
    """

    authentication_classes = ()
    renderer_classes = (PublicKeyPEMRenderer,)

    @extend_schema(
        operation_id="GetPublicKey",
        responses={200: {"type": "string"}},
        tags=["ponos"],
        examples=[
            OpenApiExample(
                name="Public key response",
                response_only=True,
                media_type="application/x-pem-file",
                status_codes=["200"],
                value=dedent(
                    """
                    -----BEGIN PUBLIC KEY-----
                    MHYwEAYHKoZIzj0CAQYFK4EEACIDYgAEmK2L6lwGzSVZwFSo0eR1z4XV6jJwjeWK
                    YCiPKdMcQnn6u5J016k9U8xZm6XyFnmgvkhnC3wreGBTFzwLCLZCD+F3vo5x8ivz
                    aTgNWsA3WFlqjSIEGz+PAVHSNMobBaJm
                    -----END PUBLIC KEY-----
                    """
                ),
            )
        ],
    )
    def get(self, request, *args, **kwargs):
        return Response(load_private_key().public_key())


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveTaskFromAgent",
        description=dedent("""
        Retrieve a Ponos task.

        Requires **guest** access on the task's process, Ponos agent authentication, or authentication as the Ponos task to retrieve.
        """),
    ),
    put=extend_schema(
        operation_id="UpdateTaskFromAgent",
        description=dedent("""
        Update a task.

        Requires authentication as the Ponos agent assigned to the task.
        """),
    ),
    patch=extend_schema(
        operation_id="PartialUpdateTaskFromAgent",
        description=dedent("""
        Partially update a task.

        Requires authentication as the Ponos agent assigned to the task.
        """),
    ),
)
class TaskDetailsFromAgent(RetrieveUpdateAPIView):

    # Avoid stale read when a recently assigned agent wants to update
    # the state of one of its tasks
    queryset = Task.objects.using("default").select_related(
        # Serialized in responses
        'agent__farm',
        'gpu',
        # Used for permission checks
        'process__corpus',
        'process__revision__repo',
    )
    authentication_classes = (
        TaskAuthentication,
        AgentAuthentication,
        TokenAuthentication,
        SessionAuthentication,
    )
    permission_classes = (
        # On all HTTP methods, require either any Ponos agent, an instance admin, the task itself, or guest access to the process' task
        IsAgentOrTaskGuest,
        # On unsafe HTTP methods, require a Ponos agent assigned to the task. Both permission classes are combined.
        IsAssignedAgentOrReadOnly,
    )
    serializer_class = TaskSerializer


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateAgent",
        tags=["ponos"],
    ),
)
class AgentRegister(CreateAPIView):
    """
    Perform agent registration and authentication.
    """
    authentication_classes = ()
    serializer_class = AgentCreateSerializer

    def get_object(self):
        if not hasattr(self.request, "data") or "public_key" not in self.request.data:
            return
        key_hash = uuid.UUID(
            hashlib.md5(self.request.data["public_key"].encode("utf-8")).hexdigest()
        )
        return Agent.objects.filter(id=key_hash).first()

    def get_serializer(self, *args, **kwargs):
        return super().get_serializer(self.get_object(), *args, **kwargs)


@extend_schema_view(
    post=extend_schema(
        operation_id="RefreshAgentToken",
        tags=["ponos"],
    )
)
class AgentTokenRefresh(TokenRefreshView):
    """
    Refresh a Ponos agent token when it expires
    """


@extend_schema(tags=["ponos"])
class AgentDetails(RetrieveAPIView):
    """
    Retrieve details of an agent including its running tasks

    Requires authentication with a verified e-mail address. Cannot be used with Ponos agent or task authentication.
    """
    authentication_classes = (TokenAuthentication, SessionAuthentication)
    permission_classes = (IsVerified, )
    serializer_class = AgentDetailsSerializer
    queryset = Agent.objects.select_related('farm')


@extend_schema(
    description="List the state of all Ponos agents",
    tags=["ponos"],
)
class AgentsState(ListAPIView):
    """
    List all agents on the system with their health state.

    Requires authentication with a verified e-mail address. Cannot be used with Ponos agent or task authentication.
    """
    authentication_classes = (TokenAuthentication, SessionAuthentication)
    permission_classes = (IsVerified, )
    serializer_class = AgentStateSerializer

    queryset = (
        Agent.objects.all()
        .annotate(
            running_tasks_count=Count("tasks", filter=Q(tasks__state=State.Running))
        )
        .prefetch_related("farm")
        .order_by("hostname")
    )


@extend_schema(
    tags=["ponos"],
    parameters=[
        OpenApiParameter(
            'cpu_load',
            type=float,
            description="Current CPU usage for this agent",
            required=True
        ),
        OpenApiParameter(
            'ram_load',
            type=float,
            description="Current RAM usage for this agent",
            required=True
        ),
    ]
)
class AgentActions(RetrieveAPIView):
    """
    Fetch the next actions an agent should perform.

    Requires authentication as a Ponos agent.
    """

    permission_classes = (IsAgent,)
    authentication_classes = (AgentAuthentication,)
    serializer_class = AgentActionsSerializer

    def get_object(self):
        return self.request.user

    def retrieve(self, request, *args, **kwargs):
        # Update agent load and last_ping timestamp
        errors = defaultdict(list)
        cpu_load = request.query_params.get("cpu_load")
        ram_load = request.query_params.get("ram_load")
        if not cpu_load:
            errors["cpu_load"].append("This query parameter is required.")
        if not ram_load:
            errors["ram_load"].append("This query parameter is required.")
        if errors:
            raise ValidationError(errors)
        # Handle fields validation with DRF as for a PATCH
        agent_serializer = AgentLightSerializer(
            self.request.user,
            data={"cpu_load": cpu_load, "ram_load": ram_load},
            partial=True,
        )
        agent_serializer.is_valid(raise_exception=True)
        # Update agent load and last_ping attributes
        agent_serializer.save(last_ping=timezone.now())
        # Retrieve next tasks after the DB has been updated with the new agent load
        response = super().retrieve(self, request, *args, **kwargs)
        return response


@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveTaskDefinition",
        tags=["ponos"],
    )
)
class TaskDefinition(RetrieveAPIView):
    """
    Obtain a task's definition.
    This holds all the required data to start a task, except for the artifacts.

    Requires authentication as a Ponos agent.
    """

    # We need to specify the default database to avoid stale reads
    # when a task is updated by an agent, then the agent immediately fetches its definition
    queryset = Task.objects.using("default").select_related("process")
    serializer_class = TaskDefinitionSerializer

    # This cannot be restricted to only the agent assigned to the task because agents need to access
    # the parent tasks of an assigned task to download their artifacts.
    authentication_classes = (AgentAuthentication, )
    permission_classes = (IsAgent, )


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListArtifacts",
        description=dedent("""
        List the artifacts of a task.

        Requires **guest** access on the task's process, Ponos agent authentication, or authentication as the Ponos task to list artifacts of.
        """),
    ),
    post=extend_schema(
        operation_id="CreateArtifact",
        description=dedent("""
        Create an artifact on task.

        Requires authentication as the Ponos task to create an artifact on, or as a Ponos agent assigned to the task.
        """),
    ),
)
class TaskArtifacts(ListCreateAPIView):

    # Used for OpenAPI schema serialization: the ID in the path is the task ID
    queryset = Task.objects.none()
    serializer_class = ArtifactSerializer
    permission_classes = (
        # On all HTTP methods, require either any Ponos agent, an instance admin, the task itself, or guest access to the process' task
        IsAgentOrTaskGuest,
        # On unsafe HTTP methods, require a Ponos agent assigned to the task, or the task itself. Both permission classes are combined.
        IsAssignedAgentOrTaskOrReadOnly,
    )

    # Force no pagination, even when global settings add them
    pagination_class = None

    @property
    def task(self):
        task = get_object_or_404(
            # Select the required tables for permissions checking
            Task.objects.select_related('process__corpus', 'process__revision__repo'),
            pk=self.kwargs["pk"],
        )
        self.check_object_permissions(self.request, task)
        return task

    def get_queryset(self):
        return self.task.artifacts.all()

    def perform_create(self, serializer):
        # Assign task when creating through the API
        serializer.save(task=self.task)


class TaskArtifactDownload(APIView):
    """
    Redirects to the S3 URL of an artifact in order to download it.
    """

    permission_classes = (IsAgentOrArtifactGuest, )

    def get_object(self, pk, path):
        artifact = get_object_or_404(
            # Select the required tables for permissions checking
            Artifact.objects.select_related('task__process__corpus', 'task__process__revision__repo'),
            task_id=pk,
            path=path,
        )
        self.check_object_permissions(self.request, artifact)
        return artifact

    @extend_schema(
        operation_id="DownloadArtifact",
        responses={302: None},
        tags=["ponos"],
    )
    def get(self, request, *args, **kwargs):
        artifact = self.get_object(*args, **kwargs)
        return redirect(artifact.s3_url)


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateTask",
        tags=["ponos"],
    )
)
class TaskCreate(CreateAPIView):
    """
    Create a task that depends on an existing task.

    Requires authentication as a Ponos task. Tasks can only be created on the process of the authenticated task.
    """

    authentication_classes = (TaskAuthentication, )
    permission_classes = (IsTask, )
    serializer_class = NewTaskSerializer


@extend_schema(tags=["ponos"])
@extend_schema_view(
    put=extend_schema(
        description=dedent("""
        Update a task.

        Requires **admin** access on the task's process, or to be the creator of the process.
        Cannot be used with Ponos agent or task authentication.
        """),
    ),
    patch=extend_schema(
        description=dedent("""
        Partially update a task.

        Requires **admin** access on the task's process, or to be the creator of the process.
        Cannot be used with Ponos agent or task authentication.
        """),
    ),
)
class TaskUpdate(UpdateAPIView):
    # Only allow regular users, not Ponos agents or tasks
    authentication_classes = (TokenAuthentication, SessionAuthentication)
    # Only allow regular users that have admin access to the task's process
    permission_classes = (IsTaskAdmin, )
    queryset = Task.objects.select_related('process__corpus', 'process__revision__repo')
    serializer_class = TaskTinySerializer


@extend_schema(tags=["ponos"])
class SecretDetails(RetrieveAPIView):
    """
    Retrieve a Ponos secret's content as cleartext.

    Requires authentication as a Ponos task.
    """

    authentication_classes = (TaskAuthentication, )
    permission_classes = (IsTask, )
    serializer_class = ClearTextSecretSerializer
    queryset = Secret.objects.all()
    lookup_field = 'name'


@extend_schema(tags=["ponos"])
class FarmList(ListAPIView):
    """
    List all available farms.

    Cannot be used with Ponos agent or task authentication.
    """
    serializer_class = FarmSerializer
    queryset = Farm.objects.order_by("name")
    authentication_classes = (TokenAuthentication, SessionAuthentication)
    permission_classes = (IsVerified, )
