import json
from hashlib import md5

from django.conf import settings
from django.db.models import CharField, Value
from django.db.models.functions import Cast, Concat, NullIf

from arkindex.project.tools import RTrimChr

__default_farm_id = None


def get_default_farm_id():
    """
    Return the default Ponos Farm ID.
    If defined, use PONOS_DEFAULT_FARM setting.
    Otherwise, pick the first existing farm or create it.

    This method uses a global variable to store the cached value retrieved in a idempotant way.
    It could return an invalid farm ID in case the default farm is removed or updated during runtime.
    """
    from arkindex.ponos.models import Farm

    global __default_farm_id
    if __default_farm_id is not None:
        return __default_farm_id
    if settings.PONOS_DEFAULT_FARM:
        farm = Farm.objects.get(pk=settings.PONOS_DEFAULT_FARM)
    else:
        farm = Farm.objects.first() or Farm.objects.create(name="Default farm")
    __default_farm_id = farm.id
    return __default_farm_id


def hash_object(object):
    object_json = json.dumps(object, sort_keys=True).encode('utf-8')
    return md5(object_json).hexdigest()


def annotate_image_url(field="image"):
    """
    Build the image URL by concatenating the server's URL to the image's path
    Server URLs might end in a slash, but not all the time,
    so we strip any trailing slashes and add our own.
    This uses NullIf as this concatenation will return '/' if there is no zone on the element.
    """
    return NullIf(
        Concat(
            RTrimChr(f'{field}__server__url', Value('/')),
            Value('/'),
            Cast(f'{field}__path', CharField()),
        ),
        Value('/'),
    )
