
import uuid
from collections import defaultdict
from textwrap import dedent

from django.db.models import Count, Q
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.validators import UniqueTogetherValidator

from arkindex.documents.serializers.elements import ElementListSerializer
from arkindex.ponos.models import Task
from arkindex.project.serializer_fields import EnumField
from arkindex.training.models import (
    Dataset,
    DatasetElement,
    DatasetState,
    MetricKey,
    MetricMode,
    MetricValue,
    Model,
    ModelVersion,
    ModelVersionState,
    default_sets,
)
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


def _model_from_context(serializer_field):
    if isinstance(serializer_field.parent.instance, ModelVersion):
        return serializer_field.parent.instance.model
    return serializer_field.context.get('model')


def _corpus_from_context(serializer_field):
    if isinstance(serializer_field.parent.instance, Dataset):
        return serializer_field.parent.instance.corpus
    return serializer_field.context.get('corpus')


_model_from_context.requires_context = True
_corpus_from_context.requires_context = True


class ModelLightSerializer(serializers.ModelSerializer):
    name = serializers.CharField(max_length=100)

    class Meta:
        model = Model
        fields = ('id', 'name')


class ModelSerializer(ModelLightSerializer):
    # Actually define the field to avoid the field-level automatically generated UniqueValidator
    rights = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Model
        fields = ModelLightSerializer.Meta.fields + ('created', 'updated', 'description', 'rights')

    def create(self, validated_data):
        instance = super().create(validated_data)

        # Create admin membership for creator user
        instance.memberships.create(
            user=self.context['request'].user,
            level=Role.Admin.value
        )
        return instance

    @extend_schema_field(serializers.ListField(child=serializers.ChoiceField(['read', 'write', 'admin'])))
    def get_rights(self, model):
        # to avoid a stale read on get_max_level when creating a model (because the corresponding membership is
        # created at the same time as the model) if a model is being created we return these rights without
        # actually checking them.
        if self.context['request'].method == 'POST':
            return ['read', 'write', 'admin']

        level = get_max_level(self.context['request'].user, model)
        rights = ['read']
        if level >= Role.Contributor.value:
            rights.append('write')
        if level >= Role.Admin.value:
            rights.append('admin')
        return rights


class CreateModelErrorResponseSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing model, if the error comes from a duplicate name.")
    name = serializers.CharField(required=False, help_text="Name of an existing model, if the error comes from a duplicate name.")
    detail = serializers.CharField(required=False, help_text="A generic error message when an error occurs outside of a specific field.")


class ModelVersionLightSerializer(serializers.ModelSerializer):
    """
    Serialize a simple model version
    """

    model = ModelLightSerializer(default=_model_from_context)
    state = EnumField(ModelVersionState)
    configuration = serializers.JSONField(style={'base_template': 'textarea.html'})

    class Meta:
        model = ModelVersion
        fields = ('id', 'model', 'tag', 'state', 'size', 'configuration')
        read_only_fields = ('id', 'model', 'state', 'size', 'hash', 'archive_hash')


class ModelVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a model version with fields that can be updated regardless of its state.
    """
    model = serializers.HiddenField(default=_model_from_context)
    parent = serializers.PrimaryKeyRelatedField(queryset=ModelVersion.objects.none(), default=None)
    description = serializers.CharField(required=False, style={'base_template': 'textarea.html'})
    configuration = serializers.JSONField(required=False, decoder=None, encoder=None, style={'base_template': 'textarea.html'})
    tag = serializers.CharField(allow_null=True, max_length=50, required=False, default=None)
    state = EnumField(ModelVersionState, read_only=True)
    s3_url = serializers.SerializerMethodField(read_only=True, help_text='Only returned if the user has **contributor** access to the model, and the model is available.')
    s3_put_url = serializers.SerializerMethodField(read_only=True, help_text='Only returned if the user has **contributor** access to the model, and the model is **not** available.')

    class Meta:
        model = ModelVersion
        fields = ('id', 'model', 'model_id', 'parent', 'description', 'tag', 'state', 'size', 'hash', 'configuration', 's3_url', 's3_put_url', 'created')
        read_only_fields = ('id', 'model', 'state', 'size', 'hash', 'archive_hash')
        validators = [
            UniqueTogetherValidator(
                queryset=ModelVersion.objects.filter(tag__isnull=False),
                fields=['model', 'tag'],
                message='A version for this model with this tag already exists.'
            )
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.instance, ModelVersion):
            model = self.instance.model
        else:
            model = self.context.get('model')
        if model:
            qs = ModelVersion.objects.filter(model_id=model.id)
            if getattr(self.instance, 'id', None):
                qs = qs.exclude(id=self.instance.id)
            self.fields['parent'].queryset = qs

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_put_url(self, obj):
        if not self.context.get('is_contributor') or obj.state == ModelVersionState.Available:
            return None
        return obj.s3_put_url

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_url(self, obj):
        if not self.context.get('is_contributor') or obj.state != ModelVersionState.Available:
            return None
        return obj.s3_url


class ModelVersionValidateSerializer(serializers.ModelSerializer):
    """
    A serializer used to update a model version to `available` state.
    """
    class Meta:
        model = ModelVersion
        fields = ('size', 'hash', 'archive_hash')
        extra_kwargs = {
            'size': {'required': True},
            'hash': {'required': True},
            'archive_hash': {'required': True},
        }


class MetricValueSerializer(serializers.ModelSerializer):
    class Meta:
        model = MetricValue
        fields = ('value', 'created', 'step')


class MetricValueCreateSerializer(MetricValueSerializer):
    """
    Create a new metric value, and the associated metric key if it doesn't already exist
    """
    model_version_id = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.select_related('model'),
        write_only=True
    )
    name = serializers.CharField(write_only=True)
    mode = EnumField(MetricMode, allow_null=True, required=False, write_only=True, default=MetricMode.Series)

    class Meta:
        model = MetricValue
        fields = MetricValueSerializer.Meta.fields + ('model_version_id', 'name', 'mode')

    def validate_model_version_id(self, model_version_id):
        # Assert user has a contributor access to the model
        access_level = get_max_level(self.context['request'].user, model_version_id.model)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail='You do not have contributor access to this model.')
        return model_version_id

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)
        metric_value = {}
        name = data.get('name')
        # Because the mode defaults to MetricMode.Series if it is not set in the request, we need to retrieve what was
        # originally sent (or not) in the request to validate; otherwise if a request tries to create a metric value for
        # a metric key of type point, without setting mode, then data.get('mode') would be 'series' which would trigger
        # the metric_key.mode != mode validation error.
        mode = self.initial_data.get('mode')
        step = data.get('step')
        value = data.get('value')
        model_version = data.get('model_version_id')

        # Retrieve or create the metric key.
        # Annotating the Metric Key with the count of associated Metric Values (as metric_values_count) in order to check
        # for duplicate values for point mode metrics, and only check for duplicate values on series mode metrics for the
        # same step if there are existing values for this Metric Key.
        # Also annotating with the count of values with step set for the Metric Key, in order to check that a metric value
        # with step set isn't being sent for a Metric Key that already has step-less values, and vice-versa.
        queryset = MetricKey.objects \
            .annotate(metric_values_count=Count('values')) \
            .annotate(values_with_steps=Count('values', filter=Q(values__step__isnull=False)))
        # If step is being sent in the request, annotate with the count of values already set for that metric key for that
        # step, in order to check that no duplicate is being sent.
        if step is not None:
            queryset = queryset.annotate(current_step_values=Count('values', filter=Q(values__step=step)))
        metrickey, created = queryset \
            .get_or_create(name=name, model_version_id=model_version.id, defaults={'mode': data.get('mode')})

        # Check that the user isn't trying to set `step` on a point type metric key, which doesn't make sense as steps are
        # only relevant for series type metric values.
        if metrickey.mode == MetricMode.Point and step is not None:
            errors['step'].append('The step property cannot be set for Point type metrics.')

        # Check that the user isn't trying to change the metric key's mode, or adding a value to a point type metric value
        # if the metric key already exists.
        if not created:
            if mode and mode != metrickey.mode:
                errors['mode'].append(f'Metric Key {metrickey.name} already exists and is of mode {str(metrickey.mode)}.')
            if metrickey.mode == MetricMode.Point and metrickey.metric_values_count > 0:
                errors['__all__'].append(f'Metric Key {metrickey.name} is of mode Point and already has a value.')
            if metrickey.metric_values_count > 0 and metrickey.mode == MetricMode.Series and step is not None:
                if metrickey.current_step_values > 0:
                    errors['step'].append('This metric has already been set for this step.')
            # One metric key cannot have values both with and without step set
            if metrickey.metric_values_count > 0 and metrickey.mode == MetricMode.Series:
                if metrickey.values_with_steps > 0 and step is None:
                    errors['step'].append(f'Cannot add a metric value without a step for metric key {metrickey.name} '
                                          'because it already has metric values with steps set.')
                elif metrickey.values_with_steps == 0 and step is not None:
                    errors['step'].append(f'Cannot add a metric value with step set for metric key {metrickey.name} '
                                          'because it already has metric values with no step set.')

        metric_value['metric'] = metrickey
        metric_value['step'] = step
        metric_value['value'] = value

        if errors:
            raise ValidationError(errors)

        return metric_value


class MetricValueBulkItemSerializer(serializers.Serializer):
    name = serializers.CharField(write_only=True)
    mode = EnumField(MetricMode, allow_null=True, required=False, write_only=True, default=MetricMode.Series)
    value = serializers.FloatField()


class MetricValueBulkSerializer(serializers.Serializer):
    model_version_id = serializers.PrimaryKeyRelatedField(queryset=ModelVersion.objects.select_related('model').all(), write_only=True)
    step = serializers.IntegerField(min_value=0, required=False)
    metrics = MetricValueBulkItemSerializer(
        many=True,
        allow_empty=False,
        help_text=dedent("""
        A list of metric values with their metric key name and mode.

        Only one value may be set at once for each metric key.
        """),
    )

    def validate_model_version_id(self, model_version_id):
        # Assert user has a contributor access to the model
        access_level = get_max_level(self.context['request'].user, model_version_id.model)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail='You do not have contributor access to this model.')
        return model_version_id

    def validate(self, data):
        errors = defaultdict(lambda: defaultdict(list))
        data = super().validate(data)
        model_version_id = data.get('model_version_id')
        step = data.get('step')

        # Retrieve the existing Metric Keys corresponding to the request metric names
        metric_names = set([item['name'] for item in data['metrics']])
        queryset = MetricKey.objects \
            .annotate(metric_values_count=Count('values')) \
            .annotate(values_with_steps=Count('values', filter=Q(values__step__isnull=False))) \
            .filter(name__in=metric_names, model_version=model_version_id)
        if step is not None:
            queryset = queryset.annotate(current_step_values=Count('values', filter=Q(values__step=step)))
        metrickeys = {
            item.name: item
            for item in queryset
        }

        # Validate each Metric Value
        metric_names_list = []
        for i, item in enumerate(data['metrics']):
            name = item['name']
            if name not in metric_names_list:
                metric_names_list.append(name)
            else:
                errors[i]['__all__'].append(f'Cannot send multiple values at once for the same name ({name}).')
            # Because the mode defaults to MetricMode.Series if it is not set in the request, we need to retrieve what was
            # originally sent (or not) in the request to validate; otherwise if a request tries to create a metric value for
            # a metric key of type point, without setting mode, then data['metrics'][i].get('mode') would be 'series' which
            # would trigger the metric_key.mode != mode validation error.
            mode = self.initial_data['metrics'][i].get('mode')
            metric_key = metrickeys.get(name)
            item['metric_key'] = metric_key

            # Check that 'step' isn't set when sending Point mode metric values
            if step and ((metric_key and metric_key.mode != MetricMode.Series) or (mode and mode != 'series')):
                errors[i]['step'].append('The step property cannot be set for Point type metrics.')
            # Check for incoherent mode and metric_key.mode
            if metric_key and mode and mode != str(metric_key.mode).lower():
                errors[i]['mode'].append(f'Metric Key {name} already exists and is of mode {metric_key.mode}.')
            # Only one metric value per point mode metric key
            if metric_key and metric_key.mode == MetricMode.Point and metric_key.metric_values_count > 0:
                errors[i]['__all__'].append(f'Metric Key {name} is of mode Point and already has a value.')
            # Only one metric value per step for series mode metric keys
            if step and metric_key and metric_key.mode == MetricMode.Series and metric_key.metric_values_count > 0:
                if metric_key.current_step_values > 0:
                    errors[i]['step'].append('This metric has already been set for this step.')
            # One metric key cannot have values both with and without step set
            if metric_key and metric_key.mode == MetricMode.Series and metric_key.metric_values_count > 0:
                if metric_key.values_with_steps > 0 and step is None:
                    errors[i]['step'].append(f'Cannot add a metric value without a step for metric key {metric_key.name} '
                                             'because it already has metric values with steps set.')
                elif metric_key.values_with_steps == 0 and step is not None:
                    errors[i]['step'].append(f'Cannot add a metric value with step set for metric key {metric_key.name} '
                                             'because it already has metric values with no step set.')

        if errors:
            raise ValidationError({'metrics': errors})

        return data

    def create(self, validated_data):
        # Create a list of metric keys to create
        metric_keys = []
        for item in validated_data['metrics']:
            if not item.get('metric_key'):
                metric_key = MetricKey(
                    id=uuid.uuid4(),
                    name=item['name'],
                    model_version=validated_data['model_version_id'],
                    mode=item.get('mode')
                )
                metric_keys.append(metric_key)
                item['metric_key'] = metric_key

        # Create Metric Keys
        MetricKey.objects.bulk_create(metric_keys)

        # Create Metric Values
        MetricValue.objects.bulk_create(
            MetricValue(
                id=uuid.uuid4(),
                value=item['value'],
                metric=item['metric_key'],
                step=validated_data.get('step')
            )
            for item in validated_data['metrics']
        )

        return validated_data


class DatasetLightSerializer(serializers.ModelSerializer):
    state = EnumField(DatasetState, read_only=True)
    description = serializers.CharField(required=True, style={'base_template': 'textarea.html'})
    name = serializers.CharField(max_length=100, required=True)
    sets = serializers.ListField(child=serializers.CharField(max_length=50), default=default_sets)
    corpus = serializers.HiddenField(default=_corpus_from_context)

    class Meta:
        model = Dataset
        fields = ('id', 'name', 'description', 'state', 'sets', 'corpus',)
        read_only_fields = ('id', 'corpus',)
        validators = [
            UniqueTogetherValidator(
                queryset=Dataset.objects.all(),
                fields=['corpus', 'name'],
                message='A dataset with this name already exists in this corpus.'
            )
        ]

    def validate_sets(self, sets):
        if sets is not None and len(sets) < 1:
            raise ValidationError('Either do not specify set names to use the default values, or specify a non-empty list of names.')
        if sets is not None and len(set(sets)) != len(sets):
            raise ValidationError('Set names must be unique.')
        return sets

    def validate(self, data):
        data = super().validate(data)
        if self.context['request'].method == 'POST':
            data['corpus_id'] = self.context['corpus'].id
            data['creator_id'] = self.context['request'].user.id
        return data


class DatasetSerializer(DatasetLightSerializer):
    state = EnumField(
        DatasetState,
        required=False,
        help_text=dedent("""
        State can only be updated through Ponos task authentication from a process containing this dataset
        and is limited to these transitions:

            Open → Building → Complete
                      ↕
                    Error

        When updated to Complete, the dataset is automatically linked to the task used for authentication.
        """),
    )
    corpus = serializers.HiddenField(default=_corpus_from_context)
    creator = serializers.CharField(
        source='creator.display_name',
        read_only=True,
        help_text='Display name of the user who created the dataset.',
    )

    def validate_state(self, state):
        """
        Dataset's state update is limited to these transitions:
        Open → Building → Complete
                  ↕
                Error
        """
        if not isinstance(self.instance, Dataset) or state == self.instance.state:
            return state

        transitions = {
            DatasetState.Open: (DatasetState.Building,),
            DatasetState.Building: (DatasetState.Complete, DatasetState.Error),
            DatasetState.Error: (DatasetState.Building,)
        }
        if state not in transitions.get(self.instance.state, ()):
            raise ValidationError(f'Transition from {self.instance.state} to {state} is not allowed.')
        return state

    def validate(self, data):
        data = super().validate(data)
        if not isinstance(self.instance, Dataset):
            return data

        state = data.get('state')
        if state is None or state == self.instance.state:
            return data

        # Dataset's state update requires a Ponos task authentication
        request = self.context.get('request')
        if request and not isinstance(request.auth, Task):
            raise ValidationError({
                'state': ['Ponos task authentication is required to update the state of a Dataset.']
            })
        # Dataset's state update is only allowed on tasks of Dataset processes, that have this dataset included
        if not request.auth.process.datasets.filter(id=self.instance.id).exists():
            raise ValidationError({'state': ['A task can only update the state of one of the datasets of its process.']})
        # Link a completed dataset to the current task which generated its artifacts
        if state == DatasetState.Complete:
            data['task_id'] = request.auth.id

        return data

    class Meta(DatasetLightSerializer.Meta):
        fields = ('id', 'name', 'description', 'sets', 'state', 'corpus_id', 'corpus', 'creator', 'task_id', 'created', 'updated')
        read_only_fields = ('id', 'corpus_id', 'corpus', 'creator', 'task_id', 'created', 'updated')


class DatasetElementSerializer(serializers.ModelSerializer):
    element = ElementListSerializer(read_only=True, allow_null=False)

    class Meta:
        model = DatasetElement
        fields = ('set', 'element')
        read_only_fields = fields


class ElementDatasetSerializer(serializers.ModelSerializer):
    dataset_name = serializers.CharField(max_length=100, source='dataset__name')

    class Meta:
        model = DatasetElement
        fields = ('dataset_id', 'dataset_name', 'set')


class SelectionDatasetElementSerializer(serializers.Serializer):
    dataset_id = serializers.PrimaryKeyRelatedField(
        queryset=Dataset.objects.all(),
        source='dataset',
        write_only=True,
        help_text="UUID of a dataset to add elements from your corpus' selection.",
        style={'base_template': 'input.html'},
    )
    set = serializers.CharField(
        max_length=50,
        write_only=True,
        help_text='Name of the set elements will be added to.',
    )

    def validate_dataset_id(self, dataset):
        if (
            (corpus := self.context.get('corpus'))
            and dataset.corpus_id != corpus.id
        ):
            raise ValidationError(f'Dataset {dataset.id} is not part of corpus {corpus.name}.')
        if dataset.state == DatasetState.Complete:
            raise ValidationError(f'Dataset {dataset.id} is marked as completed.')
        return dataset

    def validate(self, data):
        data = super().validate(data)
        dataset = data['dataset']
        if data['set'] not in dataset.sets:
            raise ValidationError({'set': [f'This dataset only allows one of {", ".join(dataset.sets)}.']})
        return data

    def create(self, validated_data):
        user = self.context['request'].user
        corpus = self.context['corpus']

        DatasetElement.objects.bulk_create(
            (
                DatasetElement(element_id=elt_id, **validated_data)
                for elt_id in (
                    user.selected_elements
                    .filter(corpus=corpus)
                    .values_list("id", flat=True)
                )
            ),
            ignore_conflicts=True
        )
        return validated_data
